/*
     data_select( Opts ) :-
     Randomly select a number of data/n items from a file.
     A number for each class label is selected.
     
     Opts: sel*(Class,Number), to(OutFile), from(InFile).

     2008:
     upsh data_select from=pk_582.pl sel=atm=Active,58 sel=atm=Inactive,58 sgm=10

     lupsh data_select from=PyruvateKinase_3D_50k_nerr.pl sel=atm=Inactive,582 to=pyruvate_kinase_inactive_582.pl

     lupsh data_select from=PyruvateKinase_3D_50k_nerr.pl sel=atm=Inactive,1164 sel=atm=Active,582 to=pk_582vs1164.pl

     lupsh data_select from= pk_582vs1164.pl sel=atm=Active,528 sel=atm=Inactive,1056 to=pk_528vs1056.pl rem=pk_54vs108.pl
*/

:- ensure_loaded( library(ordsets) ). % ord_union/3.
:- ensure_loaded( library(lists) ).   % append/3, member/2, memberchk/2.
:- ensure_loaded( library(random) ).  % randseq/3.
:- ensure_loaded( library(system) ).  % shell/1.

:- requires( sieve_indices/4 ).
:- requires( line_indexed_sieve_terms_file/3 ).
:- requires( to_number/2 ).
:- requires( nth_replace/5 ).
:- requires( opts_have_help/1 ).
:- requires( write_headed_columns/2 ).
:- requires( write_def/2 ).
:- requires( nls/1 ).
:- requires( required_opt/4 ).
:- requires( required_opt/5 ).
:- requires( defaulted_opt/4 ).
:- requires( dot/0 ).
:- requires( atoms_concat/2 ).
:- requires( min_two_digit_number_codes/2 ).

data_select( Opts ) :-
     opts_have_help( Opts ),
     !,
     data_select_help.

data_select( Opts ) :-
     data_select_defaults( Defs ), 
     append( Opts, Defs, All ),
     Goal = data_select/2,
     Sel  = sel(Class,Number), 
     required_opt( Goal, Sel, Class-Number, All, Pairs ),
     write( pairs(Pairs) ), nl,
     required_opt( Goal, from, All, [InFile|_] ),
     defaulted_opt( Goal, data_arity, All, DArs ),
     write( dars(DArs) ), nl,
     DArs = [DArity|_],
     % memberchk( data_arity(DArity), All ),
     % mold_empty_lists( Pairs, Counters ),
     memberchk( sgm(Sgms),All ),
     integer( Sgms ), Sgms > 0,
     findall( [], member(_,Pairs), Emptys ),
     data_select_segments( Sgms, Pairs, InFile, DArity, Emptys, Emptys, All ).

/*
     data_select_segments( I, Seen, All ) :- 
          Seen is a list of equal length to Pairs.
          Each element holds the list of elements seen for the
          corresponding class in Pairs.
*/
data_select_segments( 0, _Pairs, _InF, _DAr, _Seen, _Emptys, _All ) :- !.
data_select_segments( Sgm, Pairs, InF, DAr, Seen, Emptys, All ) :-
     write( Sgm ), nl,
     open( InF, read, In ),
     read( In, Fst ),

     read_in_class_counters( Fst, In, Pairs, Seen, DAr, 1, Emptys, OutCounters ),
     select_class_counters( Pairs, Seen, OutCounters, [], NxSeen, SelCounters ),
     sort( SelCounters, SelIndices ),
     close( In ),
     memberchk( to(ToTkn), All ),
     number_codes( Sgm, SgmCs ),
     min_two_digit_number_codes( SgmCs, TwoCs ),
     atom_codes( Two, TwoCs ),
     ( atom_concat(Stem,'.pl',InF) -> true ; Stem = InF ),
     underscore_non_empty_token( ToTkn, UnsToTkn ),
     atoms_concat( [Stem,'_',Two,UnsToTkn,'.pl'], OutFile ),
     memberchk( rem(RemTkn), All ),
     underscore_non_empty_token( RemTkn, UnsRemTkn ),
     atoms_concat( [Stem,'_',Two,UnsRemTkn,'.pl'], RemFile ),
     line_indexed_sieve_terms_file( InF, OutFile, RemFile, SelIndices ),
     NxSgm is Sgm - 1,
     data_select_segments( NxSgm, Pairs, InF, DAr, NxSeen, Emptys, All ).

underscore_non_empty_token( '', '' ) :- !.
underscore_non_empty_token( Oth, UnsOth ) :- 
     atom_concat( '_', Oth, UnsOth ).

data_select_defaults( [data_arity(2),rem(''),sgm(1),to(test)] ).

select_class_counters( [], [], _, Counters, [], Counters ).
select_class_counters( [_Class-Number|T], [HSeen|TSeen], [H|TClassCnts], Acc, [FSeen|RSeen], SelCounters ) :-
     length( H, Length ),
     % Max is Length,
     to_number( Number, ToNumber ),
     randseq( ToNumber, Length, Indices ),
     sort( Indices, SortedIndices ),
     sort( H, SortedH ),
     sieve_indices( SortedIndices, SortedH, _, SelH ),
     ord_union( SelH, HSeen, FSeen ),
     ord_union( SelH, Acc, NxtAcc ),
     select_class_counters( T, TSeen, TClassCnts, NxtAcc, RSeen, SelCounters ).

read_in_class_counters( end_of_file, _In, _Pairs, _Seen, _Dar, _I, Counters, OutCounters ) :-
     !,
     Counters = OutCounters.
read_in_class_counters( Datum, In, Pairs, Seen, Dar, I, Counters, OutCounters ) :-
     ( Dar == n ->
          Datum =.. [data,Args]
          ;    % assume 2
          Datum =.. [data|Args]
     ),
     last( Args, Class ),
     ( nth(N,Pairs,Class-_) ->
          nth(N,Seen,NSeen),
          ( memberchk(I,NSeen) ->
               NxtCounters = Counters 
               ;
               nth_replace(N,Counters,[I|Nth],Nth,NxtCounters)
          )
          ;
          NxtCounters = Counters
     ),
     NxtI is I + 1,
     read( In, NxtDatum ),
     read_in_class_counters( NxtDatum, In, Pairs, Seen, Dar, NxtI, NxtCounters, OutCounters ).

/*
mold_empty_lists( [], [] ).
mold_empty_lists( [_|T], [[]|M] ) :-
     mold_empty_lists( T, M ).
     */

data_select_help :-
     data_select_defaults( Defs ), 
     shell( clear ),
     Head = 'data_select(Opts) :- ',  
     Clmns = [
          'Randomly select a number of terms for each given class',
          'from a file of data terms.'
          ],
     write_headed_columns( Head, Clmns ),
     OptsHead = '  Options: ',
     DOp = [dot],
     OptsClmns = [
          'data_arity(Dar)      The arity of the data,'-
                              write_def(data_arity,Defs,DOp),
          'from+(InFile)        The input file,' - write_def(from,Defs,DOp),
          'rem(RemFile)         Remainings file,' - write_def(rem,Defs,DOp),
          'sel*+(Class,Num)     How many (Num) to collect for Class,'
                                - write_def(sel,Defs,DOp),
          'sgm(Sgm)             Number of Segments'- write_def(sgm,Defs,DOp),
          'to(OutFile)          Output file, (def: <InFile>_{sel/Sgm}.pl)', '.'
          ],
     write_headed_columns( OptsHead, OptsClmns, 0, 3 ).
