
:- module( matrix_aux, [matrix_print/1,
                        matrix_multiplication/2, matrix_multiplication/3, 
                        matrix_get_columns/2, 
                        matrix_get_column/3, matrix_get_row/3,
                        matrix_copy/2, matrix_copy/3, matrix_set_list/2,
                        matrix_apply_expr/3, matrix_apply/2,
                        matrix_apply_row/3,
                        matrix_aux_version/1
                        ] ).


:- ensure_loaded( library(matrix) ).

matrix_aux_version( 0:0:0 ).

/* predicates that build on the Yap matrix library. 
*/

/* PART I- interface predicates */ 

/* 
     highly unsofisticated, only works for 2 dimensional matrices, 
     and it doesn't use any tabbing to pretty print things 
     */

matrix_print( Mtx ) :-
     % matrix_dims( Mtx, Dims ),
     % write( dims(Dims) ), nl,
     matrix_dims( Mtx, [Rows,Cols] ),
     ColsMinus is Cols - 1,
     matrix_print_rows( Rows, 0, Mtx, ColsMinus ),
     nl.

/*
example :- 
     matrix_new( ints, [3,2], [1,2,3,4,5,6], M1 ),
     matrix_new( ints, [2,2], [2,4,6,8], M2 ),
     matrix_print( M1 ),
     matrix_print( M2 ),
     trace,
     matrix_multiplication( M1, M2 ),
     matrix_print( M1 ),
     matrix_print( M2 ), 
     matrix_multiplication( M1, M2, M3 ),
     matrix_print( M3 ).
*/

matrix_multiplication( Mat1, Mat2 ) :-
     matrix_dims( Mat1, [_R1,C1] ),
     matrix_dims( Mat2, [R2,C2] ),
     C1 =:= R2,
     C2 =:= R2,   % B must be square for A = A * B
     matrix_multiplication( Mat1, Mat2, Mat1 ).

matrix_multiplication( Mat1, Mat2, Times ) :-
     matrix_dims( Mat1, [R1,C1] ),
     matrix_dims( Mat2, [R2,C2] ),
     C1 =:= R2,
     ( var(Times) ->
          matrix_type( Mat1, Type ),
          matrix_new( Type, [R1,C2], Times )
          ;
          true
     ),
     matrix_get_columns( Mat2, Clms2 ),
     matrix_multiply_rows( R1, C1, Clms2, Mat1, Mat2, Times ).

% Get all the columns of matrix: as a list of lists...
%
matrix_get_columns( Mat, Clms ) :-
     matrix_dims( Mat, [R,C] ),
     matrix_get_columns( C, 0, R, Mat, Clms ).

matrix_get_column( Mat, C, Clm ) :-
     matrix_dims( Mat, [R,_Cs] ),
     matrix_get_column( R, 0, C, Mat, Clm ).

matrix_get_row( Mat, R, Row ) :-
     matrix_dims( Mat, [_Rs,C] ),
     matrix_get_row( C, 0, R, Mat, Row ).

matrix_copy( Mat, New ) :-
     matrix_copy( Mat, _Type, New ).

matrix_copy( Mat, Type, New ) :-
     ( var(New) -> 
          ( var(Type) -> matrix_type(Mat,Type); true ),
          matrix_dims( Mat, Dims ),
          matrix_to_list( Mat, Values ),
          matrix_new( Type, Dims, Values, New )
          ;
          matrix_to_list( Mat, Values ),
          matrix_set_list( New, Values )
     ).

/* 
     example :-
          matrix_new( ints, [3,3], Mat ),
          matrix_print( Mat ),
          findall( K, (between(2,10,J),K is J * 2), Ks ),
          matrix_set_list( Mat, Ks ),
          matrix_print( Mat ).

     */
matrix_set_list( Mat, Values ) :-
     matrix_dims( Mat, [Rs,Cs] ), 
     matrix_set_positions( Rs, 0, Cs, Mat, Values ).

matrix_apply( Mat, Goal ) :-
     matrix_size( Mat, Size ), 
     Lim is Size - 1,
     (  between(0,Lim,Off),
        matrix_offset_to_arg(Mat,Off,Pos),
        matrix_get(Mat,Pos,Val),
        call( Goal, Val, New ),
        matrix_set(Mat,Pos,New),
        fail
        ;
        true
     ).

/*   apply an arithmetic expression to all elements of a matrix...
     example :- 
          matrix_new( ints, [3,3], Mat ), 
          matrix_set( Mat, [0,0], 1 ), matrix_set( Mat, [0,1], 2 ), matrix_set( Mat, [0,2], 3 ),
          matrix_set( Mat, [1,0], 4 ), matrix_set( Mat, [1,1], 5 ), matrix_set( Mat, [1,2], 6 ),
          matrix_set( Mat, [2,0], 7 ), matrix_set( Mat, [2,1], 8 ), matrix_set( Mat, [2,2], 9 ),
          matrix_print( Mat ),
          matrix_apply_expr( Mat, (1-V), V ), 
          matrix_print( Mat ).
     */
matrix_apply_expr( Mat, Expr, Var ) :-
     matrix_size( Mat, Size ), 
     Lim is Size - 1,
     (  between(0,Lim,Off),
        matrix_offset_to_arg(Mat,Off,Pos),
        matrix_get(Mat,Pos,Val),
        Val = Var,
        New is Expr,
        matrix_set(Mat,Pos,New),
        fail
        ;
        true
     ).

matrix_apply_row( Mat, Goal, PerRow ) :-
     % matrix_get_row( Mat, R, Row ) :-
     matrix_dims( Mat, [Rs,Cs] ),
     findall( RowRes, 
          (    between( 1, Rs, R ),
               matrix_get_row( Cs, 0, R, Mat, Row ),
               call( Goal, Row, RowRes )
          ),
                    PerRow ).

/* PART II- internal predicates */ 

/* 2d printing */
matrix_print_rows( 0, _Rows, _Mtx, _ColsM ) :- !.
matrix_print_rows( Rem, I, Mtx, ColsM ) :-
     matrix_get( Mtx, [I,0], Fst ),
     write( Fst ),
     matrix_print_cols( ColsM, 1, I, Mtx ),
     NxI is I + 1,
     Still is Rem - 1,
     matrix_print_rows( Still, NxI, Mtx, ColsM ).
     
matrix_print_cols( 0, _Col, _Row, _Mtx ) :- !, nl.
matrix_print_cols( RemCol, Col, Row, Mtx ) :-
     write( ',' ),
     matrix_get( Mtx, [Row,Col], Fst ),
     write( Fst ),
     NxtCol is Col + 1,
     Still is RemCol - 1,
     matrix_print_cols( Still, NxtCol, Row, Mtx ).

/* 2d MxM multiplication */
matrix_multiply_rows( 0, _C1, _Clms2, _M1, _M2, _Tms ) :- !.
matrix_multiply_rows( R1, C1, Clms2, M1, M2, Tms ) :-
     NxtR1 is R1 - 1,   % as rows are 0...(RowDim-1)
     matrix_get_row( C1, 0, NxtR1, M1, Row1 ),
     %% write( matrix_get_row( C1, 0, R1, M1, Row1 ) ), nl,
     matrix_multiply_columns( Clms2, 0, NxtR1, Row1, Tms ),
     matrix_multiply_rows( NxtR1, C1, Clms2, M1, M2, Tms ).

matrix_multiply_columns( [], _C2plus, _R1, _Row1, _Tms ).
matrix_multiply_columns( [Clm|Clms], I, R, Row, Tms ) :-
     sum_of_product( Row, Clm, 0, Sum ),
     %% write( sum_of_product( Row, Clm, 0, Sum ) ), nl,
     matrix_set( Tms, [R,I], Sum ),
     NxI is I + 1,
     matrix_multiply_columns( Clms, NxI, R, Row, Tms ).

sum_of_product( [], [], Sum, Sum ).
sum_of_product( [H1|T1], [H2|T2], Acc, Sum ) :-
     Nxt is Acc + (H1 * H2),
     sum_of_product( T1, T2, Nxt, Sum ).

matrix_get_columns( 0, _C, _R, _Mat, [] ) :- !.
matrix_get_columns( C, I, R, Mat, [Clm|Clms] ) :-
     matrix_get_column( R, 0, I, Mat, Clm ),
     NxC is C - 1,
     NxI is I + 1,
     matrix_get_columns( NxC, NxI, R, Mat, Clms ).

matrix_get_column( 0, _R, _C, _Mat, [] ) :- !.
matrix_get_column( I, R, C, Mat, [E|Clmn] ) :-
     matrix_get( Mat, [R,C], E ),
     NxI is I - 1,
     NxR is R + 1,
     matrix_get_column( NxI, NxR, C, Mat, Clmn ).

matrix_get_row( 0, _C, _R, _Mat, [] ) :- !.
matrix_get_row( C, I, R, Mat, [E|Row] ) :-
     matrix_get( Mat, [R,I], E ),
     NxC is C - 1,
     NxI is I + 1,
     matrix_get_row( NxC, NxI, R, Mat, Row ).

matrix_set_positions( 0, _RsMinus, _Cs, _Mat, [] ) :- !.
matrix_set_positions( Rs, I, Cs, Mat, Values ) :-
     matrix_set_clm_positions( Cs, 0, I, Values, Mat, RemValues ),
     RemRs is Rs - 1, 
     NxI is I + 1,
     matrix_set_positions( RemRs, NxI, Cs, Mat, RemValues ).

matrix_set_clm_positions( 0, _CsMinus, _I, Values, _Mat, Values ) :- !.
matrix_set_clm_positions( Cs, J, I, [V|Vs], Mat, RemValues ) :-
     matrix_set( Mat, [I,J], V ),
     RemCs is Cs - 1,
     NxJ is J + 1,
     matrix_set_clm_positions( RemCs, NxJ, I, Vs, Mat, RemValues ).
     
