:- use_module(library(ordsets)).
:- use_module(library(lists)).

:- op( 100, xfx, '<-' ).

% for debugging
% indep(a,b).
% indep(e,f).

% Build a BN with an accumulator

/*---------------------------------------------------------------------

InVars - variables in InBN
InDeps - dependencies as redundant ordered list [a-[b,c],b-[a,c],c-[a,b]]
InBn - BN so far as a list of directed edges

---------------------------------------------------------------------*/

/*---------------------------------------------------------------------
  Predicate: k(+Vars,-BN)
       Args: Construct BN from Vars
   Succeeds: as often as there are legal BNs
   Comments: Each BN respects the ordering in Vars, and any indep/2 constraints
---------------------------------------------------------------------*/

k(Vars,BN) :-
	findall(X-[X],member(X,Vars),Deps),
	k(Vars,[],Deps,[],BN).

/*---------------------------------------------------------------------
  Predicate: k(+Vars,+PotParents,+Deps,+InBN,-OutBN)
       Args: Construct OutBN from InBN by adding links to Vars from PotParents
   Succeeds: as often as there are legal ways to do this
   Comments: Deps of form [a-[a,b],b-[a,b]] listing all dependencies in InBN
             only update dependencies *after* constructing all edges to Child
---------------------------------------------------------------------*/
k([],_PotParents,_Deps,OutBN,OutBN).
k([Child|RVs],PotParents,Deps,InBN,OutBN) :-
	edges(PotParents,Child,Deps,InBN,[],MidBN,NewDeps),
	update_deps(NewDeps,Deps,MidDeps),
	k(RVs,[Child|PotParents],MidDeps,MidBN,OutBN).
        
	
/*---------------------------------------------------------------------
  Predicate: edges(+PotParents,+Child,+Deps,+InBN,+InNewDeps,-OutBN,-OutNewDeps)
       Args: 
   Succeeds: 
   Comments: Deps stays constant and is used to check indep constraints
             Newly introduced dependencies are stored with InNewDeps
---------------------------------------------------------------------*/
edges([],_Child,_Deps,OutBN,OutNewDeps,OutBN,OutNewDeps).           % finished
edges([Pa|Pas],Child,Deps,InBN,InNewDeps,OutBN,OutNewDeps) :-       % connect
	add_child(Child,Pa,Deps,NewDeps),
	append(NewDeps,InNewDeps,MidNewDeps),
        edges(Pas,Child,Deps,[Child<-Pa|InBN],MidNewDeps,OutBN,OutNewDeps).
edges([_Pa|Pas],Child,Deps,InBN,InNewDeps,OutBN,OutNewDeps) :-   % don't connect
        edges(Pas,Child,Deps,InBN,InNewDeps,OutBN,OutNewDeps).

update_deps([],Deps,Deps).
update_deps([X-Y|T],Deps,MidDeps) :-
	select(X-XDeps,Deps,Rest),
	ord_add_element(XDeps,Y,XDeps2),
	select(Y-YDeps,[X-XDeps2|Rest],Rest2),
	ord_add_element(YDeps,X,YDeps2),
	update_deps(T,[Y-YDeps2|Rest2],MidDeps).

update_deps_bi( [], Deps, Deps ).
update_deps_bi( [H|T], InDeps, OutDeps ) :-
	update_deps_bi_1( H, InDeps, MidDeps ),
	update_deps_bi( T, MidDeps, OutDeps ).

update_deps_bi_1([],Deps,Deps).
update_deps_bi_1([X-Y|T],Deps,MidDeps) :-
	select(X-XDeps,Deps,Rest),
	ord_add_element(XDeps,Y,XDeps2),
	select(Y-YDeps,[X-XDeps2|Rest],Rest2),
	ord_add_element(YDeps,X,YDeps2),
	update_deps_bi_1(T,[Y-YDeps2|Rest2],MidDeps).



% Deps include self-dependency
add_child_bi(Child,Parent,Deps,NewDeps) :-
	memberchk(Child-CDeps,Deps),
	memberchk(Parent-PDeps,Deps),
	get_newdeps(PDeps,CDeps,[],NewDeps).

get_newdeps([],_CDeps,Out,Out).
get_newdeps([PDep|PDeps],CDeps,In,Out) :-
	get_newdeps2(CDeps,PDep,In,Mid),
	get_newdeps(PDeps,CDeps,Mid,Out).

get_newdeps2([],_PDep,Out,Out).
get_newdeps2([CDep|CDeps],PDep,In,Out) :-
	\+ independent(PDep,CDep),
	get_newdeps2(CDeps,PDep,[PDep-CDep|In],Out).

independent(X,Y) :-
	indep(X,Y).
independent(X,Y) :-
	indep(Y,X).
