function [samples, mode, ips]=sparse_combinatorial_inference(D,initial_model,a,b,T,k_0,varargin)

% Implements "Sparse Combinatorial Inference" as described in manuscript by
% Mukherjee, Pelech, Neve, Spellman, Gray and Speed, submitted to Bioinformatics
% 
% [samples, mode, posterior_over_inputs]=sparse_combinatorial_inference(D,initial_model,alpha,beta,T,k_0,OPTIONS)
% 
% Outputs:
% samples - MCMC samples. samples.inputs{t} are the inputs in the t'th
%   sampled model; samples.LP(t) corresponding (unnormalized) log
%   probability; samples.accept(t) is 1 if proposal was accepted
% mode - single highest scoring model (MAP estimate)
% posterior_over_inputs - posterior distribution over inputs
% 
% Inputs:
% D - binary data with observations in columns, and dimensions in rows,
%   plus the last row containing responses Y; matrix is thus of size (p+1) by n
% initial_model - starting model for MCMC, should be encoded as binary row
%   vector of length p, with 1s indicating inputs which are part of model
% alpha,beta - beta hyper-parameters
% T - desired number of MCMC samples
% k_0 - sparsity hyper-parameter, see Eq. 8 in manuscript for details
% 
% OPTIONS:
% 'lambda' - prior strength parameter, see Eq. 8 in manuscript for details,
%   set to 3 by default
% 'burn_in' - MCMC burn in, set to T/10 by default
% 'k_max' - sparsity hyper-parameter, set to floor(log2(size(D,2))) by
%   default, see Eq. 8 in manuscript for details
% 
% Example:
% initial_model=zeros(1,p)  % encodes "empty" model with no inputs included
% a = 0.9; b = 0.9;  % beta hyper-parameters
% T=1000; k_0 = floor(log2(size(D,2)/15)); 
% [samples mode posteriors] = sparse_combinatorial_inference_mcmc(D,zeros(1,p),a,b,n_samples,k_0);

lambda=3;
burn_in=ceil(T/10)
k_max=floor(log2(size(D,2)))

args = varargin;
nargs = length(args);
for i=1:2:nargs
    switch args{i},
        case 'lambda',   lambda = args{i+1};
        case 'burn_in',  burn_in = args{i+1};
        case 'k_max',  k_max = args{i+1};    
        otherwise disp('Unknown option.');disp(args{i});
    end
end



silent=false;

samples.inputs=cell(1,0);
samples.LP=[];
samples.accept=[];

current_model=initial_model;
current_LL=marginal_likelihood_internal(D,initial_model,a,b);
current_LP=current_LL + min(0,k_0-sum(initial_model))*lambda;

samples.inputs{1} = find(current_model==1); 
samples.LP=current_LP;
samples.accept=1;

counter_old=0;
for i=2:T
    [new_model H_factor]= propose_model(current_model,k_max);
    new_LL = marginal_likelihood_internal(D,new_model,a,b);
    new_LP=new_LL + min(0,k_0-sum(new_model))*lambda;

    
    alpha = exp(new_LP-current_LP)*H_factor;
    
    R=min(alpha,1);
    accept=0;
    if rand<R
        current_model=new_model;
        current_LP=new_LP;
        accept=1;
    end

    samples.inputs{i} = find(current_model==1);
    samples.LP=[samples.LP current_LP];
    samples.accept=[samples.accept accept];
    
    if ~silent
        counter_new=ceil(i/T*100);
        if rem(counter_new,10)==0 & counter_old~=counter_new
            nn=0;
            for jj=1:length(samples.LP)
                nn=nn+length(samples.inputs{jj});
            end
            nn=nn/length(samples.LP);
            disp(nn)
            fprintf('..%d %%...', counter_new); counter_old=counter_new;
        end
    end

end
mean(samples.accept)

[yy ii]=sort(samples.LP,'descend');
mode=zeros(1,size(D,1)-1);
mode(samples.inputs{ii(1)})=1;

ips=compute_posterior_over_predictors(D,samples,'burn_in',burn_in);



%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Propose model (see Eq. 10 in manuscript)
%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [out H_factor]=propose_model(model,k_max)

out=model;
p=length(model);
if sum(out)<k_max
    idx=rand_int(1,p,1);
    out(idx)=~out(idx);
else
    temp=find(out==1);
    idx=rand_int(1,length(temp),1);
    out(temp(idx))=0;
end

H_factor=1;
if sum(out)==k_max
    H_factor=p/k_max;
end

if sum(model)==k_max
    H_factor=k_max/p;
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Marginal likelihood (Eq. 7 in manuscript)
%%%%%%%%%%%%%%%%%%%%%%%%%%%

function LPs=marginal_likelihood_internal(D,models,a,b)

n=size(D,2);
n_models=size(models,1);
LPs=zeros(1,n_models);

k=gammaln(a+b)-gammaln(a)-gammaln(b);

for j=1:n_models
    model=models(j,:);

    n_parents=sum(model);
    Dt=D([find(model==1) size(D,1)],:);
    n_configs=2^n_parents;
    LP=0;

    for i=1:n_configs
        n_bin=dec2bin(i-1); vec_bin=num2vec(n_bin);
        vec_bin=pad_with_zeros(vec_bin,n_parents)';

        if n_parents>0
            idx=colfind(vec_bin,Dt(1:n_parents,:));
        else
            idx=1:n;
        end
        
        n=length(idx);
        x=sum(Dt(n_parents+1,idx));
        LP= LP + gammaln(x+a) + gammaln(n-x+b) - gammaln(a+b+n) + k;
    end
    LPs(j)=LP;
end



function vec_bin=pad_with_zeros(vec_bin,n_digits)
temp=n_digits-length(vec_bin);

if temp>0
    vec_bin=[zeros(1,temp) vec_bin];
end